# $Copyright: Copyright (c) 2013 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
# $RCSfile: XIVMirrorAgent.pm,v $ $Revision: 1.1.2.19 $

eval 'exec /opt/VRTSperl/bin/perl -I `pwd`/../../lib -S $0 ${1+"$@"}'
	if 0;

package XIVMirrorAgent;
use ArrayRep;
@ISA = ("ArrayRep");

use strict;
use warnings;
use ag_i18n_inc;
use HADRSLA;
use Data::Dumper;
use File::Spec;
use File::Basename;
use Date::Parse;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %default);

@EXPORT = qw(
);
$VERSION = '1.00';

# Replication Role
use constant ROLE_MASTER => 0;
use constant ROLE_SLAVE => 1;
use constant ROLE_UNKNOWN => -1;
use constant ROLE_INVALID => -2;

# Replication Mode
use constant MODE_SYNC => 0;
use constant MODE_ASYNC => 1;
use constant MODE_UNKNOWN => -1;
use constant MODE_INVALID => -2;

# Connection state
use constant LINK_UP => 0;
use constant LINK_DOWN => 1;
use constant LINK_UNKNOWN => -1;
use constant LINK_INVALID => -2;

# Replication state
use constant REP_ACTIVE => 0;
use constant REP_INACTIVE => 1;
use constant REP_UNKNOWN => -1;
use constant REP_INVALID => -2;

# failover_required() return codes
use constant FR_NO => 0;
use constant FR_YES => 1;
use constant FR_UNKNOWN => -1;

# failover() return codes
use constant F_SUCCESS => 100;
use constant F_FAILURE => 101;

# General return codes
use constant SUCCESS => 0;
use constant ERROR => 1;

my $DEVNULL = "/dev/null";
my $passwd_replaced_str="";
my $EncPword = "";
my $DecPword = "";
my $UseProtect = 0;

my $OSTYPE = "";
if ($^O =~ /aix/i ){
	$OSTYPE = "AIX";
} elsif ($^O =~ /hpux/i ){
	$OSTYPE = "HPUX";
}elsif ($^O =~ /linux/i ){
	$OSTYPE = "LINUX";
}elsif ($^O =~ /solaris/i ){
	$OSTYPE = "SOLARIS";
}elsif ($^O =~ /MSWin32/i ){
	$OSTYPE = "WINDOWS";
	$DEVNULL = "nul";
}

%default = (
	resname => '',
	xclipath	=> '',
	status  => '',
	splittakeover => '',
	autotakeover => '',
	localxivips => '',
	remotexivips => '',
	userlocal => '',
	pwdlocal => '',
	userremote => '',
	pwdremote => '',
	role=> '',
	mode => '',
	link_state => '',
	sync_state => '',
	rep_state => '',
	vids => {},
	resource_type => 'XIVMirror',
	isconsistgrp => '',
	grpname => '',
	remote_cgname => '',
	last_replicated_snapshot_time => '',
	specified_rpo => '',
);

# function to replace the password in xcli options with # befor logging

sub replace_passwd_for_log_msg {
	my $str="";
	$str=shift;
	$str=~s/\-p\s+\w+/\-p *****/g;
	return $str;
}

sub new {
	my($proto, $resname, $href) = @_;
	my $class = ref($proto) || $proto;
	my $self = { %default };
	$self->{'resname'} = $resname;
	$self->{'xclipath'} = $href->{'XCLIPath'};
	$self->{'autotakeover'} = $href->{'AutoTakeover'};
	$self->{'splittakeover'} = $href->{'SplitTakeover'};
	$self->{'userlocal'} = $href->{'UserLocal'};
	$self->{'pwdlocal'} = $href->{'PwdLocal'};
	$self->{'userremote'} = $href->{'UserRemote'};
	$self->{'pwdremote'} = $href->{'PwdRemote'};
	$self->{'isconsistgrp'} = $href->{'IsConsistGrp'};
	$self->{'grpname'} = $href->{'GrpName'};
	#space separated ip list for local XIV
	$self->{'localxivips'} = $href->{'LocalXIVIPs'};
	#space separated ip list for remote XIV
	$self->{'remotexivips'} = $href->{'RemoteXIVIPs'};
	#space separated ip list for Local vol names
	my @volnames = split('\s+', $href->{'LocalVolNames'});
	for (my $i = 0; $i <= $#volnames; $i++) {
		my $src_vid = $volnames[$i];
		$self->{'vids'}->{$src_vid} = '';
	}

	$self->{'start_time'} = $^T;
	$self->{'role'} = ROLE_UNKNOWN;
	$self->{'mode'} = MODE_UNKNOWN;
	$self->{'link_state'} = LINK_UNKNOWN;
	$self->{'rep_state'} = REP_UNKNOWN;
	$self->{'sync_state'} = '';
	$self->{'remote_role'} = ROLE_UNKNOWN;
	$self->{'remote_sync_state'} = '';
	$self->{'remote_cgname'} = '';
	$self->{'last_replicated_snapshot_time'} = '';
	$self->{'specified_rpo'} = '';

	$self->{'xcliLocal'} = $self->{'xclipath'} . " -x" . $self->{'localxivips'}; 

	if($self->{'userlocal'} ne '' && $self->{'pwdlocal'} ne '') {
	if ($OSTYPE eq "WINDOWS" ){
		$DecPword = $self->{'pwdlocal'};
	} else {
		$EncPword=$self->{'pwdlocal'};
		$EncPword = VCSAG_SQUEEZE_SPACES($EncPword);
		$DecPword = `$ag_i18n_inc::HADECRYPT $EncPword`;

		if ($? != 0) {
		VCSAG_LOG_MSG ( "W"," Encountered errors while decrypting specified password for local array! Agent will ignore specified user/password while running xcli on local array",34);
		$DecPword="";
		$UseProtect=1;
		}
	}
		if ($UseProtect == 0 ) {
		$self->{'xcliLocal'} = $self->{'xcliLocal'} . " -u $self->{'userlocal'} -p $DecPword";
		}
	}
	$UseProtect = 0;
	
	$passwd_replaced_str=replace_passwd_for_log_msg($self->{'xcliLocal'});
	VCSAG_LOGDBG_MSG(1, "The local xcli $passwd_replaced_str");

	$self->{xcliRemote} = $self->{'xclipath'} . " -x" . $self->{'remotexivips'}; 

	if($self->{'userremote'} ne '' && $self->{'pwdremote'} ne '') {
	$EncPword="";
	$DecPword="";
	if ($OSTYPE eq "WINDOWS" ){
		$DecPword = $self->{'pwdremote'};
	} else {
		$EncPword = $self->{'pwdremote'};  
		$EncPword = VCSAG_SQUEEZE_SPACES($EncPword);
		$DecPword = `$ag_i18n_inc::HADECRYPT $EncPword`;

		if ($? != 0) {
		VCSAG_LOG_MSG ( "W"," Encountered errors while decrypting specified password for remote array! Agent will ignore specified user/password while running xcli on remote array",35);
		$DecPword="";
		$UseProtect=1;
		}
	}

		if ($UseProtect == 0 ) {
		$self->{'xcliRemote'} = $self->{'xcliRemote'} . " -u $self->{'userremote'} -p $DecPword";
		}
	}

	$passwd_replaced_str=replace_passwd_for_log_msg($self->{'xcliRemote'});
	VCSAG_LOGDBG_MSG(1, "The remote xcli $passwd_replaced_str");

	bless($self, $class);
	return $self;
}

sub DESTROY {}


sub validate_args {
	my $href = shift;
	my $ret = SUCCESS; 
	
	my $xclipath;
	($ret, $xclipath) = VCSAG_GET_ATTR_VALUE ("XCliPath", -1, 1, @ARGV);
	if ($ret != $VCSAG_SUCCESS)	{
		VCSAG_LOG_MSG("E", "The attribute XCLIPath is not configured properly", 1);
		return ERROR;
	}

	VCSAG_LOGDBG_MSG(1, "The attribute XCLIPath is $xclipath", $xclipath); 

	if($OSTYPE eq "WINDOWS") {
		$href->{'XCLIPath'} = Win32::GetShortPathName($xclipath) ;
	} else {
		$href->{'XCLIPath'} = $xclipath;
	}

	VCSAG_LOGDBG_MSG(1, "The value of XCLIPath is $href->{'XCLIPath'}", $href->{'XCLIPath'}); 

	my $localxivips;
	my $i = 1;
	while ($ret == $VCSAG_SUCCESS) {
		my $localxivip;
		($ret, $localxivip) = VCSAG_GET_ATTR_VALUE ("LocalXIVIPs", -1, $i++, @ARGV);

		if($ret == $VCSAG_SUCCESS) {
			$localxivips = $localxivips . " -m $localxivip";
		}
	}
	VCSAG_LOGDBG_MSG(1, "The attribute LocalIVIPs is $localxivips"); 
	if ($ret != $VCSAG_SUCCESS && $i == 2)	{
		VCSAG_LOG_MSG("E", "The attribute LocalXIVIPs is not configured properly", 2);
		return ERROR;
	}

	$href->{"LocalXIVIPs"} = $localxivips;

	my $remotexivips;
	$i = 1;
	$ret = $VCSAG_SUCCESS;
	while ($ret == $VCSAG_SUCCESS) {
		my $remotexivip;
		($ret, $remotexivip) = VCSAG_GET_ATTR_VALUE ("RemoteXIVIPs", -1, $i++, @ARGV);

		if($ret == $VCSAG_SUCCESS) {
			$remotexivips = $remotexivips . " -m $remotexivip";
		}
	}
	VCSAG_LOGDBG_MSG(1, "The attribute RemoteXIVIPs is $remotexivips"); 

	if ($ret != $VCSAG_SUCCESS && $i == 2)	{
		VCSAG_LOG_MSG("E", "The attribute RemoteXIVIPs is not configured properly", 3);
		return ERROR;
	}
	
	$href->{"RemoteXIVIPs"} = $remotexivips ;

	my $isconsistgrp;
	($ret, $isconsistgrp) = VCSAG_GET_ATTR_VALUE ("IsConsistencyGroup", -1, 1, @ARGV);
	if ($ret != $VCSAG_SUCCESS)	{
		VCSAG_LOG_MSG("E", "The attribute IsConsistencyGroup is not configured properly", 6);
		return ERROR;
	}
	$href->{"IsConsistGrp"} = $isconsistgrp;

	my $grpname;
	if($isconsistgrp) {
		($ret, $grpname) = VCSAG_GET_ATTR_VALUE ("GroupName", -1, 1, @ARGV);
		if ($ret != $VCSAG_SUCCESS  || $grpname eq "")	{
			VCSAG_LOG_MSG("E", "The attribute GroupName is not configured properly", 6);
			return ERROR;
		}
	}
	$href->{"GrpName"} = $grpname;

	my $localvolnames="";
	if(!$isconsistgrp) {
		$ret = $VCSAG_SUCCESS;
		$i = 1;
		while ($ret == $VCSAG_SUCCESS) {
			my $localvolname;
			($ret, $localvolname) = VCSAG_GET_ATTR_VALUE ("LocalVolNames", -1, $i++, @ARGV);

			if($ret == $VCSAG_SUCCESS) {
				$localvolnames = $localvolnames . " " . $localvolname;
			}
		}
		VCSAG_LOGDBG_MSG(1, "The attribute LocalVolNames is $localvolnames"); 

		if ($ret != $VCSAG_SUCCESS && $i == 2)	{
			VCSAG_LOG_MSG("E", "The attribute LocalVolNames is not configured properly", 4);
			return ERROR;
		}
		$localvolnames=~s/^\s+//;
	}
	$href->{"LocalVolNames"} = $localvolnames ;

	my $autotakeover;
	($ret, $autotakeover) = VCSAG_GET_ATTR_VALUE ("AutoTakeover", -1, 1, @ARGV);
	if ($ret != $VCSAG_SUCCESS)	{
		VCSAG_LOG_MSG("E", "The attribute AutoTakeover is not configured properly", 5);
		return ERROR;
	}
	$href->{"AutoTakeover"} = $autotakeover ;

	my $splittakeover;
	($ret, $splittakeover) = VCSAG_GET_ATTR_VALUE ("SplitTakeover", -1, 1, @ARGV);
	if ($ret != $VCSAG_SUCCESS)	{
		VCSAG_LOG_MSG("E", "The attribute SplitTakeover is not configured properly", 6);
		return ERROR;
	}
	$href->{"SplitTakeover"} = $splittakeover ;
	
	my $userlocal = '';
	($ret, $userlocal) = VCSAG_GET_ATTR_VALUE ("UserLocal", -1, 1, @ARGV);
	$href->{"UserLocal"} = $userlocal ;
	VCSAG_LOGDBG_MSG(1, "The attribute UserLocal is  $userlocal",  $userlocal); 

	my $pwdlocal = '';
	($ret, $pwdlocal) = VCSAG_GET_ATTR_VALUE ("PasswordLocal", -1, 1, @ARGV);
	$href->{"PwdLocal"} = $pwdlocal;

	my $userremote = '';
	($ret, $userremote) = VCSAG_GET_ATTR_VALUE ("UserRemote", -1, 1, @ARGV);
	$href->{"UserRemote"} = $userremote ;
	VCSAG_LOGDBG_MSG(1, "The attribute UserRemote is  $userremote",  $userremote); 

	my $pwdremote = '';
	($ret, $pwdremote) = VCSAG_GET_ATTR_VALUE ("PasswordRemote", -1, 1, @ARGV);
	$href->{"PwdRemote"} = $pwdremote;

	return SUCCESS;
}

sub get_error_msg {
	my $error = shift;
	$error =~ s/<DESCRIPTION>//;
	$error =~ s/<\/DESCRIPTION>//;
	$error =~ s/^\s+//;
	$error =~ s/\s+$//;
	return $error;
}

sub get_mirror_info {
	my $self = shift;
	my $mirror = shift;

	my $role = ROLE_UNKNOWN;
	my $sync_type = MODE_UNKNOWN;
	my $target_mirror = "Unknown";
	my $sync_state = "Unknown";
	my $connected = LINK_UNKNOWN;
	my $remote_role = ROLE_UNKNOWN;
	my $active_state = "REP_UNKNOWN";
	my $remote_active_state = "Unknown";
	my $remote_sync_state = "Unknown";
	my $last_replicated_snapshot_time = "Unknown";
	my $specified_rpo = "Unknown";

	if (not defined $mirror or $mirror eq "") {
		VCSAG_LOGDBG_MSG(1, "get_mirror_info: got empty volume or cg name"); 
		return ($role, $target_mirror, $sync_type, $sync_state, $connected, $active_state, $remote_role, $remote_sync_state, $last_replicated_snapshot_time, $specified_rpo);
	}

	my $mirrorlistcmd;
	if($self->{'isconsistgrp'}) {
		$mirrorlistcmd = $self->{'xcliLocal'} . " mirror_list" . " cg=$mirror";
	} else {
		$mirrorlistcmd = $self->{'xcliLocal'} . " mirror_list" . " vol=$mirror";
	}
	
	$passwd_replaced_str=replace_passwd_for_log_msg($mirrorlistcmd);
	VCSAG_LOGDBG_MSG(1, "get_mirror_info: Mirror List command $passwd_replaced_str"); 

	my $ret = open(INPUT, "$mirrorlistcmd |") or VCSAG_LOG_MSG("E", "Cannot execute xcli command: $passwd_replaced_str: $!", 7, $passwd_replaced_str, $!) and die "Command failed:$passwd_replaced_str :$!";
	my $error = "";

	my $successcheckdone = 0;
	my $error_msg = 0;
	while (<INPUT>) {
		if ((($_ !~ /XCLIRETURN STATUS="SUCCESS"/) && ($successcheckdone == 0)) || ( $error_msg == 1 )) {
			$error_msg=1;
			if ( $_ =~ /Missing user/) {
				$error = get_error_msg($_);
				VCSAG_LOG_MSG("E", "The xcli command [$passwd_replaced_str] returned error [$error]", 7, $passwd_replaced_str, $error);
				return ($role, $target_mirror, $sync_type, $sync_state, $connected, $active_state, $remote_role, $remote_sync_state, $last_replicated_snapshot_time, $specified_rpo);
			}
			if ( $_ =~ /<DESCRIPTION>/) {
				$error = get_error_msg($_);
				VCSAG_LOG_MSG("E", "The xcli command [$passwd_replaced_str] returned error [$error]", 7, $passwd_replaced_str, $error);
				return ($role, $target_mirror, $sync_type, $sync_state, $connected, $active_state, $remote_role, $remote_sync_state, $last_replicated_snapshot_time, $specified_rpo);
			}
			
			next;
		} else {
			$successcheckdone = 1;
			#mirror_list command returns SUCCESS when invalid cg or volume name provided. And gives errror in description
			#hence special handling is needed.
			if($_ =~ /<DESCRIPTION>No Mirroring definitions match the given criteria/) {
				$error = "No Mirroring definitions match the given criteria";
				VCSAG_LOG_MSG("E", "The xcli command [$passwd_replaced_str] returned error [$error]", 7, $passwd_replaced_str, $error);
				return ($role, $target_mirror, $sync_type, $sync_state, $connected, $active_state, $remote_role, $remote_sync_state, $last_replicated_snapshot_time, $specified_rpo);
			}
		}

		if (/current_role value="(.+)"/) {
			if($1 eq "Master") {
				$role = ROLE_MASTER;
			}
			elsif($1 eq "Slave") {
				$role = ROLE_SLAVE;
			}
		}
		if (/sync_type value="(.+)"/) { 
			if ( $1 =~ "async" ) {
				$sync_type = MODE_ASYNC;
			} else {
				$sync_type = MODE_SYNC;
			}
		}
		if (/remote_peer_name value="(.+)"/) { 
			$target_mirror = $1;
		}
		if (/last_replicated_snapshot_time value="(.+)"/) { 
			$last_replicated_snapshot_time = $1;
		}
		if (/specified_rpo value="(.+)"/) {
			$specified_rpo=$1
		}
		if (/sync_state value="(.+)"/) {
			$sync_state=$1
		}
		if (/connected value="(.+)"/) {
			if ( $1 eq "yes" ) {
				$connected = LINK_UP;
			} elsif ( $1 eq "no" ) {
				$connected = LINK_DOWN;
			}	
		}
		if (/active value="(.+)"/) {
			if ( $1 eq "yes" ) {
				$active_state = REP_ACTIVE;
			} elsif ( $1 eq "no" ) {
				$active_state = REP_INACTIVE;
			}	
		}
	}

	VCSAG_LOGDBG_MSG(1, "get_mirror_info: Mirror List command return value : $ret error : $error");
	close(INPUT);
	if($error ne "") {
		my @temp = split('\"', $error);			
		VCSAG_LOG_MSG("E", "The xcli command [$passwd_replaced_str] returned error [$temp[1]]", 7, $passwd_replaced_str, $temp[1]);
		return ($role, $target_mirror, $sync_type, $sync_state, $connected, $active_state, $remote_role, $remote_sync_state, $last_replicated_snapshot_time,$specified_rpo);
	}

	# get information about remote volume if link is up
	if ( $connected == LINK_UP ) {
		my $rem_mirrorlistcmd;

		if($self->{'isconsistgrp'}) {
			$rem_mirrorlistcmd = $self->{'xcliRemote'} . " mirror_list " . "cg=" . $target_mirror ;
		} else {
			$rem_mirrorlistcmd = $self->{'xcliRemote'} . " mirror_list " . "vol=" . $target_mirror ;
		}

		$passwd_replaced_str=replace_passwd_for_log_msg($rem_mirrorlistcmd);
		VCSAG_LOGDBG_MSG(1, "get_mirror_info: Remote Mirror List command $passwd_replaced_str"); 
		open(rem_mirror_list, "$rem_mirrorlistcmd |") or VCSAG_LOG_MSG("E", "Cannot execute xcli command: $passwd_replaced_str: $!", 8, $passwd_replaced_str, $!) and die "Command failed:passwd_replaced_str :$!";
	
		my $successcheckdone = 0;
		$error_msg=0;
		while (<rem_mirror_list>) {

		if ((($_ !~ /XCLIRETURN STATUS="SUCCESS"/) && ($successcheckdone == 0)) || ($error_msg == 1)){
                        $error_msg=1;
                        if ( $_ =~ /<DESCRIPTION>/) {
                                $error = get_error_msg($_);
                                VCSAG_LOG_MSG("E", "The xcli command [$passwd_replaced_str] returned error [$error]", 7, $passwd_replaced_str, $error);
                                return ($role, $target_mirror, $sync_type, $sync_state, $connected, $active_state, $remote_role, $remote_sync_state, $last_replicated_snapshot_time, $specified_rpo);
                        }
                        next;
		} else {
			$successcheckdone = 1;
			#mirror_list command returns SUCCESS when invalid cg or volume name provided. And gives errror in description
			#hence special handling is needed.
			if($_ =~ /<DESCRIPTION>No Mirroring definitions match the given criteria/) {
				$error = "No Mirroring definitions match the given criteria";
				VCSAG_LOG_MSG("E", "The xcli command [$passwd_replaced_str] returned error [$error]", 7, $passwd_replaced_str, $error);
				return ($role, $target_mirror, $sync_type, $sync_state, $connected, $active_state, $remote_role, $remote_sync_state, $last_replicated_snapshot_time, $specified_rpo);
			}
		}


			if (/current_role value="(.+)"/) {
				if($1 eq "Master") {
					$remote_role = ROLE_MASTER;
				}
				elsif($1 eq "Slave") {
					$remote_role = ROLE_SLAVE;
				}
			}
			if (/sync_state value="(.+)"/) {
				$remote_sync_state=$1;
			}
		}
		close(rem_mirror_list);
		if($error ne "") {
			VCSAG_LOG_MSG("E", "The xcli command [$passwd_replaced_str returned error [$error]", 7, $passwd_replaced_str, $error);
		}

	} else {
		VCSAG_LOG_MSG("I", "Cannot get remote volume status: replication link is not up.", 9);
    	}

	return ($role, $target_mirror, $sync_type, $sync_state, $connected, $active_state, $remote_role, $remote_sync_state, $last_replicated_snapshot_time, $specified_rpo);
}

sub get_state {
	my @pairs = ("\n");
	my $self = shift;
	my $vols = "";
	my $key = "";
	my $state = "";
	
	$self->{state} = undef;

	if($self->{'isconsistgrp'}) {
		($self->{'role'}, $self->{'remote_cgname'}, $self->{'mode'}, $self->{'sync_state'}, $self->{'link_state'}, $self->{'rep_state'}, $self->{'remote_role'}, $self->{'remote_sync_state'}, $self->{'last_replicated_snapshot_time'}, $self->{'specified_rpo'}) = $self->get_mirror_info($self->{'grpname'});

		if($self->{'role'} == ROLE_UNKNOWN || $self->{'mode'} == MODE_UNKNOWN || $self->{'link_state'} == LINK_UNKNOWN
			|| $self->{'rep_state'} == REP_UNKNOWN || $self->{'sync_state'} eq "Unknown" || $self->{'remote_cgname'} eq "Unknown") {
			return ERROR;
		}

	} else {
		foreach my $key (keys %{$self->{vids}}) {

			my $src_vid = $key;
		
			my ($role, $tgt_vid, $sync_type, $sync_state, $connected, $rep_state, $remote_role, $remote_sync_state, $last_replicated_snapshot_time, $specified_rpo) = $self->get_mirror_info($src_vid);

			VCSAG_LOGDBG_MSG(1, "Volname: $src_vid Role: $role Sync Type: $sync_type Targetvolname: $tgt_vid Sync state: $sync_state, connected: $connected rep state: $rep_state");
			VCSAG_LOGDBG_MSG(1, "remote role: $remote_role remote sync state: $remote_sync_state");

			if($role == ROLE_UNKNOWN || $sync_type == MODE_UNKNOWN || $connected == LINK_UNKNOWN
				|| $rep_state == REP_UNKNOWN || $sync_state eq "Unknown" || $tgt_vid eq "Unknown") {
				return ERROR;
			}

			push (@pairs, "$src_vid:$tgt_vid");
			# set remote vol id
			if (exists $self->{vids}->{$src_vid} && $self->{vids}->{$src_vid} eq '') {
				$self->{vids}->{$src_vid} = $tgt_vid;
			} else {
				next;
			}	

			# set local vol role
			if ($self->{'role'} == ROLE_UNKNOWN) {
				$self->{'role'} = $role;
			}elsif ($role != $self->{'role'}) {
				# volumes are in an invalid mix of roles
				VCSAG_LOG_MSG("E", "The volumes are in an invalid mix of roles: @pairs", 10, "@pairs");
				$self->{'role'} = ROLE_INVALID;
			}

			# set link state
			if ($self->{'link_state'} == LINK_UNKNOWN) {
				$self->{'link_state'} = $connected;
			} elsif ($connected != $self->{'link_state'}) {
				# volumes are in an invalid mix of link states
				VCSAG_LOG_MSG("E", "The volumes are in an invalid mix of link states: @pairs", 11, "@pairs");
				$self->{'link_state'} = LINK_INVALID;
			}

			# set replication mode
			if ($self->{'mode'} == MODE_UNKNOWN) {
				$self->{'mode'} = $sync_type;
			} elsif ($sync_type != $self->{'mode'}) {
				# volumes are in an invalid mix of modes
				VCSAG_LOG_MSG("E", "The volumes are in an invalid mix of sync types: @pairs", 33, "@pairs");
				$self->{'mode'} = MODE_INVALID;
			}

			# set activation state
			if ($self->{'rep_state'} == REP_UNKNOWN) {
				$self->{'rep_state'} = $rep_state;
			} elsif ($rep_state != $self->{'rep_state'}) {
				# volumes are in an invalid mix of activation states
				VCSAG_LOG_MSG("E", "The volumes are in an invalid mix of replication activation states: @pairs", 12, "@pairs");
				$self->{'rep_state'} = REP_INVALID;
			}

			# set sync state
			if ($self->{'sync_state'} eq '') {
				$self->{'sync_state'} = $sync_state;
			} elsif ($sync_state ne $self->{'sync_state'}) {
				# volumes are in an invalid mix of sync states
				VCSAG_LOG_MSG("E", "The volumes are in an invalid mix of sync states: @pairs", 13, "@pairs");
				$self->{'sync_state'} = "Invalid";
			}

			# set remote vol role
			if ($self->{'remote_role'} == ROLE_UNKNOWN) {
				$self->{'remote_role'} = $remote_role;
			}elsif ($remote_role != $self->{'remote_role'}) {
				# volumes are in an invalid mix of states
				VCSAG_LOG_MSG("E", "The volumes are in an invalid mix of roles of remote volumes (this could also happen when no communication to remote site): @pairs", 14, "@pairs");
				$self->{'remote_role'} = ROLE_INVALID;
			}

			# set remote sync state
			if ($self->{'remote_sync_state'} eq '') {
				$self->{'remote_sync_state'} = $remote_sync_state;
			} elsif ($remote_sync_state ne $self->{'remote_sync_state'}) {
				# volumes are in an invalid mix of sync states
				$self->{'remote_sync_state'} = "Invalid";
			}


			# set remote activation state
			#unless (defined $self->{remote_active_state}) {
			#	$self->{remote_active_state} = $remote_active_state;
			#} elsif ($remote_active_state ne $self->{remote_active_state}) {
				# volumes are in an invalid mix of activation states
			#	VCSAG_LOG_MSG("E", "XIVMirror Agent: Volumes are in an invalid mix of remote activation states: @pairs", 10, "@pairs");
			#	$self->{remote_active_state} = "Invalid";
			#}
		}
	}
	return SUCCESS;
}

sub failover_required {
	my ($self) = shift;
	my $ret = FR_YES;
	
	if($self->{'role'} == ROLE_INVALID || $self->{'mode'} == MODE_INVALID 
		|| $self->{'rep_state'} == REP_INVALID || $self->{'link_state'} == LINK_INVALID) {
		VCSAG_LOGDBG_MSG(1, "failover_required: One or more mirror attributes are invalid.");
		return FR_NO;
	}
		
	if($self->{'role'} == ROLE_SLAVE)
	{
		#is link down and autotakeover set?
		if($self->{'link_state'} == LINK_DOWN)	{
			if($self->{'autotakeover'} == 0) {
				VCSAG_LOG_MSG("N", "The replication link is down and AutoTakeover attribute is set to 0; no action will be taken.", 15);
				$ret = FR_NO;
			}
			else {
				VCSAG_LOG_MSG("N", "The replication link is down and AutoTakeover attribute is set to 1; will try to takeover.", 16);
				$ret = FR_YES;
			}
		}

		#is replication inactive and splittakover set?
		if($self->{'rep_state'} == REP_INACTIVE) {
			if($self->{'splittakeover'} == 0) {
				VCSAG_LOG_MSG("N", "The replication is in inactive state and SplitTakeover attribute is set to 0; no action will be taken.", 17);
				$ret = FR_NO;
			}
			else {
				VCSAG_LOG_MSG("N", "The replication is in inactive state and SplitTakeover attribute is set to 1; will try to takeover.", 18);
				$ret = FR_YES;
			}
		}
		
		if(($self->{'link_state'} == LINK_UP) && ($self->{'rep_state'} == REP_ACTIVE)) {
			if($self->{'mode'} == MODE_ASYNC) {
				if( $self->{'sync_state'} eq "RPO OK") {
					$ret = FR_YES;
				}
				else {
					VCSAG_LOG_MSG("N", "The replication mode is asynchronous and RPO is not OK; no action will be taken.", 19);
					$ret = FR_NO;
				}
			}
			if($self->{'mode'} == MODE_SYNC) {
				if( $self->{'sync_state'} eq "Consistent") {
					$ret = FR_YES;
				}
				else {
					VCSAG_LOG_MSG("N", "The replication mode is synchronous and data is not consistent; no action will be taken.", 20);
					$ret = FR_NO;
				}
			}
		}
	}

	return $ret;
}

sub wait_for_sync {
	my $self = shift;

	my ($oto) = split('\n', `$ag_i18n_inc::HATYPE -value XIVMirror OnlineTimeout`);

	my $cur_time = time();

	my $wait_time = ($cur_time - $self->{'start_time'}) + 60;

	VCSAG_LOGDBG_MSG(1, "The time available for waiting for synchronization $wait_time");

	while ($wait_time < $oto) {
		if(!$self->mirror_has_sync_job()) {
			VCSAG_LOGDBG_MSG(1, "No mirror has sync job");
			return SUCCESS;
		}
		sleep(5);
	}
	return ERROR;
}
			
sub mirror_has_sync_job {
	my $self = shift;
	my $syncjoblistcmd = $self->{'xcliLocal'} . " sync_job_list";
	my @syncvol_list = ();

	open(INPUT, "$syncjoblistcmd |") or VCSAG_LOG_MSG("E", "Cannot execute xcli command: $syncjoblistcmd: $!", 7, $syncjoblistcmd, $!) and die "Command failed:$syncjoblistcmd :$!";

	while (<INPUT>) {
		if (/mirror_peer value="(.+)"/) {
			push(@syncvol_list, $1);
			#VCSAG_LOGDBG_MSG(1, "The volume $1 has sync job pending");
	       }
	}
	close(INPUT);

	VCSAG_LOGDBG_MSG(1, "The no of volumes having sync job $#syncvol_list");

	for(my $i =0; $i< $#syncvol_list; $i++) {
		if($self->{'isconsistgrp'}) {
			if($self->{'grpname'} =~ $syncvol_list[$i]) {
				VCSAG_LOGDBG_MSG(1, "The consistency group $self->{'grpname'} has sync job pending.");
				return 1;
			}
		} else {
			foreach my $vol (keys %{$self->{vids}}) {
				VCSAG_LOGDBG_MSG(1, "Matching $syncvol_list[$i] and $vol");

				if($vol =~ $syncvol_list[$i]) {
					VCSAG_LOGDBG_MSG(1, "The volume $vol has sync job pending.");
					return 1;
				}
			}
		}
	}
	return 0;
}

sub get_cg_vol_list {
	my $self = shift;
	my $cmd = shift;
	my $cg_vols_list_ref = shift;
	my @vol_list = ();
	my $ret = SUCCESS;

	$passwd_replaced_str=replace_passwd_for_log_msg($cmd);
	open(INPUT, "$cmd |") or VCSAG_LOG_MSG("E", "Cannot execute xcli command: $passwd_replaced_str: $!", 7, $passwd_replaced_str, $!) and die "Command failed:$passwd_replaced_str :$!";

	my $vol_name = "";
	while (<INPUT>) {
		if (/<name value="(.+)"/) {
			$vol_name = $1;
	       }
		if (/mirrored value="(.+)"/) {
			if($1 eq "yes") {
				VCSAG_LOGDBG_MSG(1, "get_cg_vol_list vol name $vol_name is mirrored.");
				push(@vol_list, $vol_name);
			}
	       }
	}
	close(INPUT);

	@{$cg_vols_list_ref} = @vol_list;

	return $ret;
}

sub clear_local_vol_reservations {
	my $self = shift;
	my $ret = SUCCESS;
    	my $reservation_clear_command = " reservation_clear";
	my @vol_list = ();

	if($self->{'isconsistgrp'}) {
		my $volcmd = $self->{'xcliLocal'} . " vol_list cg=$self->{'grpname'}";
		$ret = $self->get_cg_vol_list($volcmd, \@vol_list);
	} else {
		@vol_list = keys %{$self->{vids}};
	}

	for(my $i =0; $i< $#vol_list; $i++) {
		# clear local volumes reservation keys
		my $clearcmd = $self->{'xcliLocal'} . $reservation_clear_command . " vol=$vol_list[$i]";

		my $output = $self->exec_xcli($clearcmd);

		if ($output !~ /SUCCESS/i) {
			$ret = ERROR;	
			$passwd_replaced_str=replace_passwd_for_log_msg($clearcmd);
			VCSAG_LOG_MSG("E", "The clear local volumes reservation keys operation $passwd_replaced_str failed with error $output", 21, $passwd_replaced_str, $output);
		}
	}
	return $ret;
}

sub clear_remote_vol_reservations {
	my $self = shift;
	my $ret = SUCCESS;
    	my $reservation_clear_command = " reservation_clear";

	my @vol_list = ();

	if($self->{'isconsistgrp'}) {
		my $volcmd = $self->{xcliRemote} . " vol_list cg=$self->{'remote_cgname'}";
		$ret = $self->get_cg_vol_list($volcmd, \@vol_list);
	} else {
		@vol_list = keys %{$self->{vids}};
	}

	for(my $i =0; $i< $#vol_list; $i++) {
		# clear local volumes reservation keys
		my $clearcmd = $self->{xcliRemote} . $reservation_clear_command . " vol=$vol_list[$i] -y";

		my $output = $self->exec_xcli($clearcmd);

		if ($output !~ /SUCCESS/i) {
			$ret = ERROR;	
			$passwd_replaced_str=replace_passwd_for_log_msg($clearcmd);
			VCSAG_LOG_MSG("E", "The clear remote volumes reservation keys operation $passwd_replaced_str failed with error $output", 22, $passwd_replaced_str, $output);
		}
	}
	return $ret;
}

sub failover_switch_roles {
	my $ret = SUCCESS;
	my $self = shift;
	my @retry_vols = ();

	if($self->{'isconsistgrp'}) {
		my $failovercmd = $self->{'xcliRemote'} . " mirror_switch_roles " . "cg=$self->{'remote_cgname'} -y";
		my $output = $self->exec_xcli($failovercmd);

		if ($output !~ /SUCCESS/i) {
			if(($output =~ /MIRROR_HAS_SYNC_JOB/i) || $output =~ /MIRROR_RETRY_OPERATION/i) {
				my ($oto) = split('\n', `$ag_i18n_inc::HATYPE -value XIVMirror OnlineTimeout`);
				my $cur_time = time();
				my $time_to_sleep = $oto - ($cur_time - $self->{'start_time'}) - (5 *  $#retry_vols) - 30;

				VCSAG_LOGDBG_MSG(1, "The time available for sleep before retrying the switch roles operation $time_to_sleep");

				#wait till maximum time available before reqtrying the operation.
				if ($time_to_sleep > 0) {
					sleep($time_to_sleep);
				}

				my $output = "";
				VCSAG_LOG_MSG("I", "Retrying switch role command for consistency group $self->{'remote_cgname'}", 23, $self->{'remote_cgname'});

				$output = $self->exec_xcli($failovercmd);

				if ($output !~ /SUCCESS/i) {
					$passwd_replaced_str=replace_passwd_for_log_msg($failovercmd);
					VCSAG_LOG_MSG("E", "The switch role operation $passwd_replaced_str failed with error $output.", 23, $passwd_replaced_str, $output);
					return ERROR;
				}
			
			} else {
				$passwd_replaced_str=replace_passwd_for_log_msg($failovercmd);
				VCSAG_LOG_MSG("E", "The switch role operation $passwd_replaced_str failed with error $output.", 23, $passwd_replaced_str, $output);
				return ERROR;
			}
		}
		return $ret;
	} 
	
	foreach my $key (keys %{$self->{vids}}) {
		# perform mirror_switch_role which will change mirror direction.
		my $failovercmd = $self->{xcliRemote} . " mirror_switch_roles " . "vol=$self->{vids}->{$key} -y";
		my $output = $self->exec_xcli($failovercmd);

		if ($output !~ /SUCCESS/i) {
			if(($output =~ /MIRROR_HAS_SYNC_JOB/i) || $output =~ /MIRROR_RETRY_OPERATION/i) {		
				push(@retry_vols, $self->{vids}->{$key});
			} else {
				$passwd_replaced_str=replace_passwd_for_log_msg($failovercmd);
				VCSAG_LOG_MSG("E", "The switch role operation $passwd_replaced_str failed with error $output.", 23, $passwd_replaced_str, $output);
				return ERROR;
			}
		}
	}

	if($#retry_vols < 0) {
		return $ret;
	}

	my ($oto) = split('\n', `$ag_i18n_inc::HATYPE -value XIVMirror OnlineTimeout`);
	my $cur_time = time();
	my $time_to_sleep = $oto - ($cur_time - $self->{'start_time'}) - (5 *  $#retry_vols) - 30;

	VCSAG_LOGDBG_MSG(1, "The time available for sleep before retrying the switch roles operation $time_to_sleep");

	#wait till maximum time available before reqtrying the operation.
	if ($time_to_sleep > 0) {
		sleep($time_to_sleep);
	}

	for(my $i = 0; $i <= $#retry_vols; $i++) {
		my $output = "";
		VCSAG_LOG_MSG("I", "Retrying switch role command for volume $retry_vols[$i]", 23, $retry_vols[$i]);

		my $failovercmd = $self->{xcliRemote} . " mirror_switch_roles " . "vol=$retry_vols[$i] -y";

		$output = $self->exec_xcli($failovercmd);

		if ($output !~ /SUCCESS/i) {
			$passwd_replaced_str=replace_passwd_for_log_msg($failovercmd);
			VCSAG_LOG_MSG("E", "The switch role operation $passwd_replaced_str failed with error $output.", 23, $passwd_replaced_str, $output);
			return ERROR;
		}
	}
	return $ret;
}

sub failover_change_role {
	my $ret = SUCCESS;
	my $self = shift;

	if($self->{'isconsistgrp'}) {
		# perform mirror_switch_role which will change mirror direction.
		my $failovercmd = $self->{xcliLocal} . " mirror_change_role " . "cg=$self->{'grpname'} -y" ;
		my $output = $self->exec_xcli($failovercmd);

		if ($output !~ /SUCCESS/i) {
			$ret = ERROR;	
			$passwd_replaced_str=replace_passwd_for_log_msg($failovercmd);
			VCSAG_LOG_MSG("E", "The change role operation $passwd_replaced_str failed with error $output", 24, $passwd_replaced_str, $output);
		}
		return $ret;
	}

	foreach my $key (keys %{$self->{vids}}) {
		# perform mirror_switch_role which will change mirror direction.
		my $failovercmd = $self->{xcliLocal} . " mirror_change_role " . "vol=$key -y" ;
		my $output = $self->exec_xcli($failovercmd);

		if ($output !~ /SUCCESS/i) {
			$ret = ERROR;	
			$passwd_replaced_str=replace_passwd_for_log_msg($failovercmd);
			VCSAG_LOG_MSG("E", "The change role operation $passwd_replaced_str failed with error $output", 24, $passwd_replaced_str, $output);
		}
	}
	return $ret;
}

sub failover {
	my ($self) = shift;
	my $ret = F_FAILURE;
	if($self->{'link_state'} == LINK_DOWN || $self->{'rep_state'} == REP_INACTIVE){
		#change role
		$ret = $self->clear_local_vol_reservations();
		if ($ret != SUCCESS) {
			VCSAG_LOG_MSG("W", "Clear local volumes reservations keys did not succeed. Problems may occur which will require administrative intervention. Continuing with failover", 25 );
		}
		$ret = $self->failover_change_role();
		if ($ret != SUCCESS) {
			VCSAG_LOG_MSG("E", "Failover with change role did not succeed; administrative intervention required", 26);
			return F_FAILURE;
		} else {
			$ret = F_SUCCESS;
		}
	}


	if(($self->{'link_state'} == LINK_UP) && ($self->{'rep_state'} == REP_ACTIVE) ){
		if ((($self->{'mode'} == MODE_ASYNC) && ($self->{'sync_state'} eq "RPO OK")) ||
			(($self->{'mode'} == MODE_SYNC) && ($self->{'sync_state'} eq "Consistent"))) {

				if(($self->{'mode'} == MODE_ASYNC) && ($self->wait_for_sync() == ERROR)) {
					VCSAG_LOG_MSG("E", "Some of the mirror pairs still have sync job. The agent will not try the switch role operation for any of the mirror volumes to maintain mirror role consistency. Please retry after some time.", 25 );
					return F_FAILURE;
				}
					
				# switch roles
				$ret = $self->clear_local_vol_reservations();
				if ($ret != SUCCESS) {
					VCSAG_LOG_MSG("W", "Clear local volumes reservations keys did not succeed. Problems may occur which will require administrative intervention. Continuing with failover", 25 );
				}
				$ret = $self->clear_remote_vol_reservations();
				if ($ret != SUCCESS) {
					VCSAG_LOG_MSG("W", "Clear remote volumes reservations keys did not succeed. Problems may occur which will require administrative intervention. Continuing with failover", 27);
				}
				$ret = $self->failover_switch_roles();

				if ($ret != SUCCESS) {
					VCSAG_LOG_MSG("E", "Failover with switch role did not succeed; the volumes might be in mixed roles; administrative intervention is required", 26);
						return F_FAILURE;

					#VCSAG_LOG_MSG("N", "Failover with switch roles did not succeed. Attempting change role operation.", 28);
					# try change role if switch roles did not succeed
					#$ret = $self->failover_change_role();
					#if ($ret != SUCCESS) {
					#	VCSAG_LOG_MSG("E", "Failover with change role did not succeed; administrative intervention is required", 26);
					#	return F_FAILURE;
					#}else {
					#	$ret = F_SUCCESS;
					#}
				}else {
					$ret = F_SUCCESS;
				}

		}
	}
	return $ret;
}


sub exec_xcli {
	my ($self, $xcli_cmd) = @_;
	my $ret_status = "";

	$passwd_replaced_str=replace_passwd_for_log_msg($xcli_cmd);
	open(INPUT, "$xcli_cmd |") or VCSAG_LOG_MSG("E", "Cannot execute xcli command: $passwd_replaced_str: $!", 7, $passwd_replaced_str, $!) and die "Command failed:$passwd_replaced_str :$!";

	while (<INPUT>) {
		if (/XCLIRETURN STATUS="(.+)"/) {
			$ret_status = $1;
		}
	}
	close(INPUT);

	VCSAG_LOGDBG_MSG(1, "The return status of XCLI command [$passwd_replaced_str] is $ret_status");

	return $ret_status;
}



#####
#fix for open EP. Temporarily added util functions here. 
#Should be removed in future when util functions are available in ArrayRep.pm
#copied relavant functions from SRDFAgent.pm
#####
sub get_lockfilename {
	my ($self) = shift;
	my $typ;
	my $lock_file;
	my $lock_dir = ""; 

	if (!defined($self->{'resource_type'}) || 
		(length($self->{'resource_type'}) == 0)) {

		$typ = `$ag_i18n_inc::HARES -value $self->{'resname'} Type`;
		chomp $typ;
	} else {
		$typ = $self->{'resource_type'};
	}

	if ($^O =~ /MSWin32/i) {
		$lock_dir = $ag_i18n_inc::VCSLOG . "\\lock\\" . $typ . "\\";
	} else {
		$lock_dir = $ag_i18n_inc::VCSLOG . "/lock/" . $typ . "/";
	}

	$lock_file = $lock_dir . ".VCS_" . $typ . "_" . $self->{'resname'} . 
		"_Online_Lockfile.DO-NOT-REMOVE";

	return $lock_file;
}

sub create_lockfile {
	my ($self) = shift;
	my ($lock_file) = $self->get_lockfilename();

	my (undef, $lock_dir, undef) = fileparse($lock_file);

	if (! -d $lock_dir) {
		mkdir  $lock_dir or 
			VCSAG_LOG_MSG("E", "Could not create lock directory $lock_dir",
				1001, "$lock_dir");
	}

	open(lock_file_handle, ">$lock_file") or 
		VCSAG_LOG_MSG("E", "Could not create file: $lock_file", 1002,
			 "$lock_file");

	close(lock_file_handle);
}

sub remove_lockfile {
	my($self) = shift;
	my ($lock_file) = $self->get_lockfilename();

	if (-f "$lock_file") {
		unlink("$lock_file") or 
			VCSAG_LOG_MSG("E", "Could not delete file: $lock_file", 1003,
				 "$lock_file");
	}
}


sub exists_lockfile {
	my ($self) = shift;
	my $retval = 0;
	my ($lock_file) = $self->get_lockfilename();

	if (-f "$lock_file") {
		$retval = 1;
	}
	return $retval;
}

#
# Return 1 if a parent resource of a parent service group is ONLINE
#

sub is_parent_online {
	my ($self) = shift;
	my ($p_res, $ret, $tmpstr, $sys);
	my @p_res_list = ();
	my $is_online = 0;

	$tmpstr = `$ag_i18n_inc::HASYS -nodeid`;
	chomp $tmpstr;
	$tmpstr =~ s/^.*Node name:[\s+](.*)[\s+]Node ID:.*$/$1/ ;
	$self->{'nodename'} = $sys = $tmpstr;

	# Are there any resources that depend on this resource
	@p_res_list = $self->get_res_deps();

	if ($#p_res_list >= 0) { # Parent resources were found
		foreach $p_res (@p_res_list) {

			$ret = system("$ag_i18n_inc::HARES -wait $p_res Probed 1 "
					." -sys $sys -time 120 2>$DEVNULL");
			if ($ret == 0) {
				my ($st) = split('\n', `$ag_i18n_inc::HARES -state $p_res -sys $sys`);		
				chomp $st;
				if ($st =~ /ONLINE/) {
					$is_online = 1;
					last;
				}
			}
		}
	} else { # No parent resources
		if ($self->is_parent_sg_res_online()) {
			$is_online = 1;	
		}
	}
	return $is_online;
}

#
# Return 1 if at least one OnOff/OnOnly resource from a parent service
# group is online.
#

sub is_parent_sg_res_online {
	my ($self) = shift;
	my ($sg_name, $p_sg_name, $line, $ret);
	my @lines = ();
	my @fields = ();
	my @valid_dep_lines = ();
	my $is_online = 0;
	my $sys = $self->{'nodename'};
	
	#
	# Get parent service groups that have an online local/online global
	# dependency to our service group
	#

	$sg_name = `$ag_i18n_inc::HARES -value $self->{'resname'} Group`;
	chomp $sg_name;

	(@lines) = split('\n', `$ag_i18n_inc::HAGRP -dep $sg_name`);
	chomp @lines;

	# VCS specific code --->
	(@valid_dep_lines) = grep (/online[\s+](local|global)/, @lines );

	foreach $line (@valid_dep_lines) {
		(@fields) = split('\s+', $line);

		if ($fields[1] eq $sg_name) {
			$p_sg_name = $fields[0];
		} else {
			next;
		}

		#
		# If at least one OnOff or OnOnly probed resource in the ParentSG
		# has gone online, then the SRDF resource can be online as well
		#

		$ret = system("$ag_i18n_inc::HAGRP -wait $p_sg_name Probed 1 -sys $sys"
				 . " -time 120 2>$DEVNULL");

		if ($ret == 0) {
			my @p_on_res_list = $self->get_res_list($p_sg_name, "ONLINE");	

			#
			# At least 1 online resource was found. Note: $#X == 0
			# implies @X has 1 element
			#
			if ($#p_on_res_list >= 0){

				#
				# We found at least one ONLINE resource. If it is of type 
				# OnOff or OnOnly, we can continue to remain ONLINE
				#   
				foreach my $p_res (@p_on_res_list) {
					my $p_type = `$ag_i18n_inc::HARES -value $p_res Type`;
					chomp $p_type;
					my $p_oprn_type = `$ag_i18n_inc::HATYPE -value $p_type Operations`;
					chomp $p_oprn_type;
					if ($p_oprn_type =~ /None/i ) {
						next;
					}

					$is_online = 1;
					last;
				}
			}
		}
		last if ($is_online == 1);
	}

	return $is_online;
}

#
# Returns the names of resources that belong to a given group and
# are in particular state
#
sub get_res_list {
	my ($self) = shift;
	my ($sg_name, $res_state) = @_;
	my @retarr = ();
	my @tmparr = ();
	my ($ret, $nodename, $resname);

	$ret = run_cmd("$ag_i18n_inc::HARES -list Group=$sg_name State=$res_state", 
				\@tmparr, 1);

	if ($ret) {
		foreach (@tmparr) {
			/^#/ && next;
			($resname, $nodename)  = split;

			if ((!grep(/^${resname}$/, @retarr)) &&
				($nodename =~ /$self->{'nodename'}/)) {

				push (@retarr, $resname);
			}
		}	
	}
	return @retarr;	
}


#
# Returns the names of parent resources
#
sub get_res_deps {
	my ($self) = shift;

	my ($sg_name, $res_state) = @_;
	my @retarr = ();
	my @tmparr = ();
	my ($ret, $grpname, $p_resname, $c_resname);

	$ret = run_cmd("$ag_i18n_inc::HARES -dep $self->{'resname'}", \@tmparr, 1);

	if ($ret) {
		foreach (@tmparr) {
			/^#/ && next;
			($grpname, $p_resname, $c_resname)  = split;
		
			#
			# Note: with VCSOne, the 'resname' would be <grpname>.<resname>
			# whereas the $c_resname would only have the trailing <resname>
			# part
			#
		
			if ((!grep(/^${p_resname}$/, @retarr)) && 
				( $c_resname =~ /$self->{'resname'}$/ )) {
				push (@retarr, $p_resname);
			}
		}	
	}
	
	return @retarr;	
}


sub run_cmd {
	my ($cmd, $output_ref, $ignore_failure) = @_;
	my $retval = 1;
	@{$output_ref} = `$cmd`;

	if ($? != 0) {
		$retval = 0;
		unless (1 == $ignore_failure ) {
			VCSAG_LOG_MSG("W", "Command: \'$cmd\' failed !", 1007,
					"$cmd");
		}
	} else {
		chomp @{$output_ref};
	}
	return $retval;
}


#
# Function to compute Replication specific RPO using xcli's
#

sub Rep_Tech_RPO {
	my $ResName=shift;
	VCSAG_SET_ENVS ($ResName, 20091);
	my $USERLANG=$ENV{"LANG"};
	$ENV{"LANG"}="C";

	my %agent_args_hash = ();
	if (XIVMirrorAgent::ERROR == XIVMirrorAgent::validate_args(\%agent_args_hash) ) {
		return 0;
	}

	my $res = XIVMirrorAgent->new($ResName, \%agent_args_hash);

	my $effective_sync_state="RPO OK";
	my $effective_specified_rpo=0;
	my $effective_Replication_RPO="";
	my $temp_effective_Replication_RPO=0;
	my $snapshot_time=0;
	my $curr_time=0;
	my $do_not_compute_RPO=0;
	my $out = "";	

	if($res->{'isconsistgrp'}) {
		($res->{'role'}, $res->{'remote_cgname'}, $res->{'mode'}, $res->{'sync_state'}, $res->{'link_state'}, $res->{'rep_state'}, $res->{'remote_role'}, $res->{'remote_sync_state'}, $res->{'last_replicated_snapshot_time'}, $res->{'specified_rpo'}) = $res->get_mirror_info($res->{'grpname'});

		#
		# Compute RPO only if mode is ASYNC and Replication link is Up, Initialized
		#

		if ( $res->{'mode'} == MODE_ASYNC && $res->{'link_state'} == LINK_UP  && $res->{'sync_state'} ne "Initializing" ) {
			if ( $res->{'last_replicated_snapshot_time'} ne "" ) {
				$snapshot_time=str2time($res->{'last_replicated_snapshot_time'});
				$curr_time=time;
				$effective_Replication_RPO=HADRSLA::time_format($curr_time-$snapshot_time);
				$effective_sync_state=$res->{'sync_state'};
				$effective_specified_rpo=$res->{'specified_rpo'};
			} else {
				$out = "Rep_Tech_RPO: Replication based RPO data as reported by XIV CLI \n";
				$out = $out . "Rep_Tech_RPO: Last_Replicated_Snapshot_Time is empty for specified consistency group. Cannot compute RPO\n"; 
				return $out;
			}
		} else {
			$do_not_compute_RPO=1;
		}
	} else {

		# When individual volume names are specified, RPO should return following
		#	- "RPO Lagging" as sync_state if any one of the volumes is in that state
		#	- Maximum effective RPO value of all the configured volumes

		foreach my $key (keys %{$res->{vids}}) {
			my $src_vid = $key;
			my ($role, $tgt_vid, $sync_type, $sync_state, $connected, $rep_state, $remote_role, $remote_sync_state, $last_replicated_snapshot_time, $specified_rpo) = $res->get_mirror_info($src_vid);

			# Compute RPO only if mode is ASYNC and Replication link is Up, Initialized

			if ( $sync_type == MODE_ASYNC && $connected == LINK_UP && $sync_state ne "Initializing" ) {
				if ( $last_replicated_snapshot_time ne "" ) {
					if ( $sync_state ne $effective_sync_state && $sync_state ne "RPO OK"){
						$effective_sync_state=$sync_state;
					}
					$curr_time=time;
					$snapshot_time=str2time($last_replicated_snapshot_time);
					my $vol_effective_Replication_RPO=$curr_time-$snapshot_time;
					if ( $vol_effective_Replication_RPO > $temp_effective_Replication_RPO ) {
						$temp_effective_Replication_RPO=$vol_effective_Replication_RPO;
					}
				} else {
					$out = "Rep_Tech_RPO: Replication based RPO data as reported by XIV CLI \n";
					$out = $out . "Rep_Tech_RPO: Last_Replicated_Snapshot_Time is empty for one/more volumes. Cannot compute RPO\n"; 
					return $out;
				}
			} else {
				$do_not_compute_RPO=1;
				last;
			}
		}
			if ( $do_not_compute_RPO == 0 ) {
				$effective_Replication_RPO=HADRSLA::time_format($temp_effective_Replication_RPO);
			}
	}


	if ( $do_not_compute_RPO == 0 ) {

		$out = "Rep_Tech_RPO: Replication based RPO data as reported by XIV CLI \n";
		if($res->{'isconsistgrp'}) {
			$out = $out . "Rep_Tech_RPO: Replication Sync State = $effective_sync_state \n";
			$out = $out . "Rep_Tech_RPO: Specified RPO = $effective_specified_rpo [hh:mm:ss] \n";
			$out = $out . "Rep_Tech_RPO: Replication RPO = $effective_Replication_RPO [hh:mm:ss] [ Current_Time - Last_Replicated_Snapshot_Time ] \n";
		} else {
			$out = $out . "Rep_Tech_RPO: Replication Sync State = $effective_sync_state [Consolidated state for specified volumes] \n";
			$out = $out . "Rep_Tech_RPO: Replication RPO = $effective_Replication_RPO [hh:mm:ss] [ Max of 'Current_Time - Last_Replicated_Snapshot_Time ' for specified volumes ] \n";
		}

	} else {
		$out = "Rep_Tech_RPO: Replication based RPO data as reported by XIV CLI \n";
		$out = $out . "Rep_Tech_RPO: The Replication Mode is Sync or Replication Link is not Up/Initialized. Cannot compute RPO\n";
	}
	return $out;
}

1;
